/**
 *    Popup.java
 *    Copyright (C) 2010 New Zealand Digital Library, http://expeditee.org
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.expeditee.gui;

import org.expeditee.core.Clip;
import org.expeditee.core.Colour;
import org.expeditee.core.EnforcedClipStack.EnforcedClipKey;
import org.expeditee.core.Fill;
import org.expeditee.core.Lifetime;
import org.expeditee.core.Stroke;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.gui.PopupManager.PopupAnimator;
import org.expeditee.items.ItemUtils;

/**
 * A Custom swing pop-up in Expeditee.
 * 
 * Expeditee pop-ups can be re-used. Pop-ups are always heavyweight (i.e. uses invalidation).
 * 
 * @see {@link PopupManager}
 * 
 * @author Brook Novak
 */
public abstract class Popup
{
	//Mike says: Can we get the border for the IW to which this popup corresponds?
	// Brook says: Would be nice - but popups are actually independent from widgets
	//   =>Now: It is up to the user of the popup to set the border thickness
	private static final Stroke DEFAULT_STROKE = new Stroke(2.0f);
	
	private Stroke _borderStroke = DEFAULT_STROKE;
	private Colour _borderColour = Colour.BLACK;
	private Fill _backgroundFill = new Fill(Colour.WHITE);
	
	private PopupAnimator _animator = null;
	
	private Lifetime _autoHideTime = new Lifetime(Lifetime.INFINITE_LIFETIME);
	private Runnable _autoHideRoutine = new Runnable()
	{
		@Override
		public void run()
		{
			hide();
			DisplayController.invalidateArea(getBounds());
			DisplayController.requestRefresh(true);
		}
	};
	
	private boolean _hidden = true;

	private boolean _consumeBackClick = false;

	/**
	 * Creates a new pop-up.
	 * Auto-hide is set to true.
	 */
	public Popup(PopupAnimator animator)
	{
		setAnimator(animator);
	}
	
	public Popup(PopupAnimator animator, long lifetime)
	{
		this(animator);
		_autoHideTime.setLifetime(lifetime);
	}
	
	public Lifetime getAutoHideTime()
	{
		return _autoHideTime;
	}
	
	public void setAnimator(PopupAnimator animator)
	{
		_animator = animator;
	}
	
	public PopupAnimator getAnimator()
	{
		return _animator;
	}
	
	public boolean isShowing()
	{
		return !_hidden;
	}
	
	public void show()
	{
		if (_hidden) onShow();
		
		if (_animator != null) {
			_animator.show();
			if (!_animator.isFinished()) {
				_autoHideTime.cancelExpiry();
			} else {
				_autoHideTime.refresh();
				_autoHideTime.onExpiry(_autoHideRoutine);
			}
		} else {
			_autoHideTime.refresh();
			_autoHideTime.onExpiry(_autoHideRoutine);
		}
		
		_hidden = false;
	}

	public void hide()
	{
		if (!_hidden) onHide();
		
		if (_animator != null) _animator.hide();
		_autoHideTime.cancelExpiry();
		_hidden = true;
	}
	
	public boolean update()
	{
		if (_animator != null) {
			if (!_hidden && !_animator.isFinished()) {
				if (!_animator.update()) {
					_autoHideTime.refresh();
					_autoHideTime.onExpiry(_autoHideRoutine);
				}
			}
			
			if (_hidden && !_animator.isFinished()) {
				return _animator.update();
			}
		} else {
			if (_autoHideTime.hasExpired()) hide();
		}
		
		return false;
		
	}
	
	public final void paint()
	{
		GraphicsManager g = EcosystemManager.getGraphicsManager();
		
		// Make sure the popup isn't hidden
		AxisAlignedBoxBounds currentBounds = getBounds();
		if (currentBounds == null) return;
		
		// Draw border and background
		g.drawRectangle(currentBounds, 0.0, _backgroundFill, _borderColour, _borderStroke, null);
		
		EnforcedClipKey key = g.pushClip(new Clip(currentBounds));
		
		paintInternal();
		
		g.popClip(key);
	}
	
	/** Draws the interior paint area of the popup. */
	protected abstract void paintInternal();
	
	/** Gets the bounds of this popup when it's fully shown. */
	public abstract AxisAlignedBoxBounds getFullBounds();
	
	/** Gets the bounds of this popup in its current animated state. */
	public AxisAlignedBoxBounds getBounds()
	{
		if (_animator == null) {
			if (_autoHideTime.hasExpired()) {
				return null;
			} else {
				return getFullBounds();
			}
		}
		
		return _animator.getAnimatedBounds(getFullBounds());
	}
	
	/** Invoked when the popup becomes hidden, or when the popup is animating to show but cancelled. */
	public abstract void onHide();
	
	/** Invoked when the popup shows. Note this might not eventuate for animated popups. */
	public abstract void onShow();

	public boolean shouldConsumeBackClick()
	{
		return _consumeBackClick;
	}

	/**
	 * @param consumeBackClick
	 * 		Set to True for whenever the user clicks empty space
	 * 		to go back a frame that if this popup is visible should
	 * 		consume the back-click event.
	 */
	protected void setConsumeBackClick(boolean consumeBackClick)
	{
		_consumeBackClick = consumeBackClick;
	}
	
	/**
	 * @return
	 * 		True if this popup auto hides.
	 */
	public boolean doesAutoHide()
	{
		return _autoHideTime.remainingLifetime() != Lifetime.INFINITE_LIFETIME;
	}
	
	/**
	 * Invalidates self.
	 * 
	 * @param thickness
	 * 		The new thickness to set. Null for no border.
	 */
	public void setBorderThickness(float thickness)
	{
		assert(thickness >= 0);
		
		if (_borderStroke != null && _borderStroke.thickness == thickness) return;
		
		boolean posInvalidate = true;
		
		if (thickness < _borderStroke.thickness) {
			invalidateAppearance();
			posInvalidate = false;
		}
		
		if (thickness == 0) {
			_borderStroke = null;
		} else {
			_borderStroke = new Stroke(thickness);
		}
		
		if (posInvalidate) invalidateAppearance();

	}
	
	/**
	 * @return
	 * 		The border thickness of this popup. Zero or more.
	 */
	public float getBorderThickness()
	{
		if (_borderStroke == null) return 0.0f;

		return _borderStroke.thickness;
	}
	
	/**
	 * Sets the border color around the popup.
	 * Invalidates self.
	 * 
	 * @param c
	 * 		The new color. Null for transparent.
	 */
	public void setBorderColor(Colour c)
	{
		if (c == null && _borderColour != null)  invalidateAppearance();
		
		if (c != _borderColour) {
			_borderColour = c;
			invalidateAppearance();
		}
	}
	
	/**
	 * @return
	 * 		The border color for the popup. NUll if transparent
	 */
	public Colour getBorderColor()
	{
		return _borderColour;
	}
	
	/**
	 * Invalidates the whole pop-up so that it must be fully repainted.
	 */
	public void invalidateAppearance()
	{
		AxisAlignedBoxBounds bounds = getBounds();
		
		if (bounds == null) return;
		
		if (_borderColour != null && _borderStroke != null && _borderStroke.thickness > 0) { // border
			DisplayController.invalidateArea(ItemUtils.expandRectangle(bounds, (int)Math.ceil(getBorderThickness()) + 1));
		} else { // no border
			DisplayController.invalidateArea(bounds);
		}

	}
}
